import json
import logging
from getpass import getpass

from azure.common.credentials import ServicePrincipalCredentials, UserPassCredentials, get_azure_cli_credentials
from msrestazure.azure_active_directory import MSIAuthentication
from ScoutSuite.core.console import print_info
from msrestazure.azure_active_directory import AADTokenCredentials
import adal

from ScoutSuite.providers.base.authentication_strategy import AuthenticationStrategy, AuthenticationException


class AzureCredentials:

    def __init__(self,
                 arm_credentials, aad_graph_credentials,
                 tenant_id=None, default_subscription_id=None):

        self.arm_credentials = arm_credentials  # Azure Resource Manager API credentials
        self.aad_graph_credentials = aad_graph_credentials  # Azure AD Graph API credentials
        self.tenant_id = tenant_id
        self.default_subscription_id = default_subscription_id

    def get_tenant_id(self):
        if self.tenant_id:
            return self.tenant_id
        else:
            return self.aad_graph_credentials.token['tenant_id']


class AzureAuthenticationStrategy(AuthenticationStrategy):

    def authenticate(self,
                     cli=None, user_account=None, user_account_browser=None,
                     service_principal=None, file_auth=None, msi=None,
                     tenant_id=None,
                     subscription_id=None,
                     client_id=None, client_secret=None,
                     username=None, password=None,
                     programmatic_execution=False,
                     **kargs):
        """
        Implements authentication for the Azure provider
        """
        try:

            # Set logging level to error for libraries as otherwise generates a lot of warnings
            logging.getLogger('adal-python').setLevel(logging.ERROR)
            logging.getLogger('msrest').setLevel(logging.ERROR)
            logging.getLogger('urllib3').setLevel(logging.ERROR)

            if cli:
                arm_credentials, subscription_id, tenant_id = get_azure_cli_credentials(with_tenant=True)
                aad_graph_credentials, placeholder_1, placeholder_2 = \
                    get_azure_cli_credentials(with_tenant=True, resource='https://graph.windows.net')

            elif user_account:

                if not (username and password):
                    if not programmatic_execution:
                        username = username if username else input("Username: ")
                        password = password if password else getpass("Password: ")
                    else:
                        raise AuthenticationException('Username and/or password not set')

                arm_credentials = UserPassCredentials(username, password)
                aad_graph_credentials = UserPassCredentials(username, password,
                                                        resource='https://graph.windows.net')

            elif user_account_browser:

                # As per https://docs.microsoft.com/en-us/samples/azure-samples/data-lake-analytics-python-auth-options
                # /authenticating-your-python-application-against-azure-active-directory/
                # The client id used above is a well known that already exists for all azure services. While it makes
                # the sample code easy to use, for production code you should use generate your own client ids for
                # your application.
                client_id = '04b07795-8ddb-461a-bbee-02f9e1bf7b46'
                authority_host_uri = 'https://login.microsoftonline.com'
                authority_uri = authority_host_uri + '/' + tenant_id
                context = adal.AuthenticationContext(authority_uri, api_version=None)

                # Resource Manager
                resource_uri = 'https://management.core.windows.net/'
                code = context.acquire_user_code(resource_uri, client_id)
                print_info('To authenticate to the Resource Manager API, use a web browser to '
                           'access {} and enter the {} code.'.format(code['verification_url'],
                                                                     code['user_code']))
                mgmt_token = context.acquire_token_with_device_code(resource_uri, code, client_id)
                arm_credentials = AADTokenCredentials(mgmt_token, client_id)
                # Graph
                resource_uri = 'https://graph.windows.net'
                code = context.acquire_user_code(resource_uri, client_id)
                print_info('To authenticate to the Azure Graph API, use a web browser to '
                           'access {} and enter the {} code.'.format(code['verification_url'],
                                                                     code['user_code']))
                mgmt_token = context.acquire_token_with_device_code(resource_uri, code, client_id)
                aad_graph_credentials = AADTokenCredentials(mgmt_token, client_id)

            elif service_principal:

                if not tenant_id:
                    if not programmatic_execution:
                        tenant_id = input("Tenant ID: ")
                    else:
                        raise AuthenticationException('No Tenant ID set')

                if not client_id:
                    if not programmatic_execution:
                        client_id = input("Client ID: ")
                    else:
                        raise AuthenticationException('No Client ID set')

                if not client_secret:
                    if not programmatic_execution:
                        client_secret = getpass("Client secret: ")
                    else:
                        raise AuthenticationException('No Client Secret set')

                arm_credentials = ServicePrincipalCredentials(
                    client_id=client_id,
                    secret=client_secret,
                    tenant=tenant_id
                )

                aad_graph_credentials = ServicePrincipalCredentials(
                    client_id=client_id,
                    secret=client_secret,
                    tenant=tenant_id,
                    resource='https://graph.windows.net'
                )

            elif file_auth:

                data = json.loads(file_auth.read())
                tenant_id = data.get('tenantId')
                client_id = data.get('clientId')
                client_secret = data.get('clientSecret')

                arm_credentials = ServicePrincipalCredentials(
                    client_id=client_id,
                    secret=client_secret,
                    tenant=tenant_id
                )

                aad_graph_credentials = ServicePrincipalCredentials(
                    client_id=client_id,
                    secret=client_secret,
                    tenant=tenant_id,
                    resource='https://graph.windows.net'
                )

            elif msi:

                arm_credentials = MSIAuthentication()
                aad_graph_credentials = MSIAuthentication(resource='https://graph.windows.net')

            else:
                raise AuthenticationException('Unknown authentication method')

            return AzureCredentials(arm_credentials, aad_graph_credentials,
                                    tenant_id, subscription_id)

        except Exception as e:
            raise AuthenticationException(e)



from ScoutSuite.providers.aws.facade.base import AWSFacade
from ScoutSuite.providers.aws.resources.regions import Regions

from .keys import Keys


class KMS(Regions):
    _children = [
        (Keys, 'keys'),
    ]

    def __init__(self, facade: AWSFacade):
        super(KMS, self).__init__('kms', facade)



from ScoutSuite.providers.azure.authentication_strategy import AzureCredentials
from ScoutSuite.providers.azure.facade.aad import AADFacade
from ScoutSuite.providers.azure.facade.arm import ARMFacade
from ScoutSuite.providers.azure.facade.keyvault import KeyVaultFacade
from ScoutSuite.providers.azure.facade.network import NetworkFacade
from ScoutSuite.providers.azure.facade.securitycenter import SecurityCenterFacade
from ScoutSuite.providers.azure.facade.sqldatabase import SQLDatabaseFacade
from ScoutSuite.providers.azure.facade.storageaccounts import StorageAccountsFacade
from ScoutSuite.providers.azure.facade.virtualmachines import VirtualMachineFacade
from ScoutSuite.providers.azure.facade.appservice import AppServiceFacade

from azure.mgmt.resource import SubscriptionClient

from ScoutSuite.core.console import print_info, print_exception

# Try to import proprietary services
try:
    from ScoutSuite.providers.azure.facade.appgateway_private import AppGatewayFacade
except ImportError:
    pass
try:
    from ScoutSuite.providers.azure.facade.loadbalancer_private import LoadBalancerFacade
except ImportError:
    pass
try:
    from ScoutSuite.providers.azure.facade.rediscache_private import RedisCacheFacade
except ImportError:
    pass


class AzureFacade:
    def __init__(self,
                 credentials: AzureCredentials,
                 subscription_ids=[], all_subscriptions=False,
                 programmatic_execution=False):

        self.credentials = credentials
        self.programmatic_execution = programmatic_execution

        self.subscription_list = []
        self.subscription_ids = subscription_ids
        self.all_subscriptions = all_subscriptions

        self.aad = AADFacade(credentials)
        self.arm = ARMFacade(credentials)
        self.keyvault = KeyVaultFacade(credentials)
        self.virtualmachines = VirtualMachineFacade(credentials)
        self.network = NetworkFacade(credentials)
        self.securitycenter = SecurityCenterFacade(credentials)
        self.sqldatabase = SQLDatabaseFacade(credentials)
        self.storageaccounts = StorageAccountsFacade(credentials)
        self.appservice = AppServiceFacade(credentials)

        # Instantiate facades for proprietary services
        try:
            self.appgateway = AppGatewayFacade(credentials)
        except NameError:
            pass
        try:
            self.loadbalancer = LoadBalancerFacade(credentials)
        except NameError:
            pass
        try:
            self.rediscache = RedisCacheFacade(credentials)
        except NameError:
            pass

        self._set_subscriptions()

    async def get_subscriptions(self):
        if self.subscription_list:
            return self.subscription_list
        else:
            self._set_subscriptions()

    def _set_subscriptions(self):

        # Create the client
        subscription_client = SubscriptionClient(self.credentials.arm_credentials)
        # Get all the accessible subscriptions
        accessible_subscriptions_list = list(subscription_client.subscriptions.list())

        if not accessible_subscriptions_list:
            print_exception('The provided credentials do not have access to any subscriptions')
            self.subscription_list = []
            return

        # Final list, start empty
        subscriptions_list = []

        # No subscription provided, infer
        if not (self.subscription_ids or self.all_subscriptions):
            try:
                # Tries to read the subscription list
                print_info('No subscription set, inferring ID')
                s = next(subscription_client.subscriptions.list())
            except StopIteration:
                print_info('Unable to infer a subscription')
                # If the user cannot read subscription list, ask Subscription ID:
                if not self.programmatic_execution:
                    s = input('Subscription ID: ')
                else:
                    print_exception('Unable to infer a Subscription ID')
                    raise
            finally:
                print_info('Running against the "{}" subscription'.format(s.subscription_id))
                subscriptions_list.append(s)

        # A specific set of subscriptions
        elif self.subscription_ids:
            # Only include accessible subscriptions
            subscriptions_list = [s for s in accessible_subscriptions_list if
                                  s.subscription_id in self.subscription_ids]
            # Verbose skip
            for s in self.subscription_ids:
                if not any(subs.subscription_id == s for subs in accessible_subscriptions_list):
                    print_info('Skipping subscription "{}": this subscription does not exist or '
                               'is not accessible with the provided credentials'.format(s))
            print_info('Running against {} subscription(s)'.format(len(subscriptions_list)))

        # All subscriptions
        elif self.all_subscriptions:
            subscriptions_list = accessible_subscriptions_list
            print_info('Running against {} subscription(s)'.format(len(subscriptions_list)))

        # Other == error
        else:
            print_exception('Unknown Azure subscription option')
            self.subscription_list = []
            raise

        if subscriptions_list:
            self.subscription_list = subscriptions_list
        else:
            print_exception('No subscriptions to scan')
            self.subscription_list = []
            raise



from ScoutSuite.providers.azure.facade.base import AzureFacade
from ScoutSuite.providers.azure.resources.base import AzureResources


class ServerAzureAdAdministrators(AzureResources):

    def __init__(self, facade: AzureFacade, resource_group_name: str, server_name: str, subscription_id: str):
        super(ServerAzureAdAdministrators, self).__init__(facade)
        self.resource_group_name = resource_group_name
        self.server_name = server_name
        self.subscription_id = subscription_id

    async def fetch_all(self):
        self['ad_admin'] = await self.facade.sqldatabase.get_server_azure_ad_administrators(
            self.resource_group_name, self.server_name, self.subscription_id)



from azure.mgmt.security import SecurityCenter

from ScoutSuite.core.console import print_exception
from ScoutSuite.providers.utils import run_concurrently


class SecurityCenterFacade:

    def __init__(self, credentials):
        self.credentials = credentials

    def get_client(self, subscription_id: str):
        return SecurityCenter(self.credentials.arm_credentials, subscription_id, '')

    async def get_pricings(self, subscription_id: str):
        try:
            client = self.get_client(subscription_id)
            pricings_list = await run_concurrently(
                lambda: client.pricings.list()
            )
            if hasattr(pricings_list, 'value'):
                return pricings_list.value
            else:
                return []
        except Exception as e:
            print_exception('Failed to retrieve pricings: {}'.format(e))
            return []

    async def get_security_contacts(self, subscription_id: str):
        try:
            client = self.get_client(subscription_id)
            return await run_concurrently(
                lambda: list(client.security_contacts.list())
            )
        except Exception as e:
            print_exception('Failed to retrieve security contacts: {}'.format(e))
            return []

    async def get_auto_provisioning_settings(self, subscription_id: str):
        try:
            client = self.get_client(subscription_id)
            return await run_concurrently(
                lambda: list(client.auto_provisioning_settings.list())
            )
        except Exception as e:
            print_exception('Failed to retrieve auto provisioning settings: {}'.format(e))
            return []

    async def get_information_protection_policies(self, subscription_id: str):
        try:
            client = self.get_client(subscription_id)
            scope = '/subscriptions/{}'.format(self._subscription_id)
            return await run_concurrently(lambda: list(client.information_protection_policies.list(scope=scope)))
        except Exception as e:
            print_exception('Failed to retrieve information protection policies: {}'.format(e))
            return []

    async def get_settings(self, subscription_id: str):
        try:
            client = self.get_client(subscription_id)
            return await run_concurrently(
                lambda: list(client.settings.list())
            )
        except Exception as e:
            print_exception('Failed to retrieve settings: {}'.format(e))
            return []

    async def get_alerts(self, subscription_id: str):
        try:
            client = self.get_client(subscription_id)
            return await run_concurrently(
                lambda: list(client.alerts.list())
            )
        except Exception as e:
            print_exception('Failed to retrieve alerts: {}'.format(e))
            return []

    async def get_compliance_results(self, subscription_id: str):
        try:
            client = self.get_client(subscription_id)
            scope = '/subscriptions/{}'.format(subscription_id)
            return await run_concurrently(
                lambda: list(client.compliance_results.list(scope=scope))
            )
        except Exception as e:
            print_exception('Failed to retrieve compliance results: {}'.format(e))
            return []

    async def get_regulatory_compliance_results(self, subscription_id: str):
        try:
            client = self.get_client(subscription_id)
            results = []
            try:
                compliance_standards = await run_concurrently(
                    lambda: list(client.regulatory_compliance_standards.list())
                )
            except Exception as e:
                print_exception('Failed to retrieve regulatory compliance standards: {}'.format(e))
                return {}
            else:
                for standard in compliance_standards:
                    try:
                        compliance_controls = await run_concurrently(
                            lambda: list(client.regulatory_compliance_controls.list(regulatory_compliance_standard_name=standard.name))
                        )
                        for control in compliance_controls:
                            control.standard_name = standard.name
                            results.append(control)
                    except Exception as e:
                        print_exception('Failed to retrieve compliance controls: {}'.format(e))
                        pass
            finally:
                return results
        except Exception as e:
            print_exception('Failed to retrieve regulatory compliance results: {}'.format(e))
            return []





from ScoutSuite.providers.aws.facade.base import AWSFacade
from ScoutSuite.providers.aws.resources.base import AWSResources


class Aliases(AWSResources):
    def __init__(self, facade: AWSFacade, region: str):
        super(Aliases, self).__init__(facade)
        self.region = region

    async def fetch_all(self):
        raw_aliases = await self.facade.kms.get_aliases(self.region)
        for raw_alias in raw_aliases:
            id, alias = self._parse_alias(raw_alias)
            self[id] = alias

    def _parse_alias(self, raw_alias):
        alias_dict = {
            # all KMS Aliases are prefixed with alias/, so we'll strip that off
            'name': raw_alias.get('AliasName').lstrip('alias/'),
            'arn': raw_alias.get('AliasArn'),
            'key_id': raw_alias.get('TargetKeyId')}
        return alias_dict['name'], alias_dict



from azure.mgmt.authorization import AuthorizationManagementClient

from ScoutSuite.core.console import print_exception
from ScoutSuite.providers.utils import run_concurrently


class ARMFacade:
    def __init__(self, credentials):
        self.credentials = credentials

    def get_client(self, subscription_id: str):
        return AuthorizationManagementClient(self.credentials.arm_credentials, subscription_id=subscription_id)

    async def get_roles(self, subscription_id: str):
        try:
            client = self.get_client(subscription_id)
            scope = '/subscriptions/{}'.format(subscription_id)
            return await run_concurrently(lambda: list(client.role_definitions.list(scope=scope)))
        except Exception as e:
            print_exception('Failed to retrieve roles: {}'.format(e))
            return []

    async def get_role_assignments(self, subscription_id: str):
        try:
            client = self.get_client(subscription_id)
            return await run_concurrently(lambda: list(client.role_assignments.list()))
        except Exception as e:
            print_exception('Failed to retrieve role assignments: {}'.format(e))
            return []



from ScoutSuite.providers.azure.facade.base import AzureFacade
from ScoutSuite.providers.azure.resources.base import AzureResources
from ScoutSuite.providers.utils import get_non_provider_id


class WebApplication(AzureResources):

    def __init__(self, facade: AzureFacade, subscription_id: str):
        super(WebApplication, self).__init__(facade)
        self.subscription_id = subscription_id

    async def fetch_all(self):
        for raw_web_app in await self.facade.appservice.get_web_apps(self.subscription_id):
            id, web_app = self._parse_web_app(raw_web_app)
            self[id] = web_app

    def _parse_web_app(self, raw_web_app):

        web_app_dict = {}
        web_app_dict['id'] = get_non_provider_id(raw_web_app.id)
        web_app_dict['name'] = raw_web_app.name
        web_app_dict['kind'] = raw_web_app.kind
        web_app_dict['location'] = raw_web_app.location
        web_app_dict['type'] = raw_web_app.type
        web_app_dict['tags'] = raw_web_app.tags
        web_app_dict['state'] = raw_web_app.state
        web_app_dict['host_names'] = raw_web_app.host_names
        web_app_dict['repository_site_name'] = raw_web_app.repository_site_name
        web_app_dict['usage_state'] = raw_web_app.usage_state
        web_app_dict['enabled'] = raw_web_app.enabled
        web_app_dict['https_only'] = raw_web_app.https_only
        web_app_dict['enabled_host_names'] = raw_web_app.enabled_host_names
        web_app_dict['availability_state'] = raw_web_app.availability_state
        web_app_dict['host_name_ssl_states'] = raw_web_app.host_name_ssl_states
        web_app_dict['server_farm_id'] = raw_web_app.server_farm_id
        web_app_dict['reserved'] = raw_web_app.reserved
        web_app_dict['is_xenon'] = raw_web_app.is_xenon
        web_app_dict['hyper_v'] = raw_web_app.hyper_v
        web_app_dict['last_modified_time_utc'] = raw_web_app.last_modified_time_utc
        web_app_dict['site_config'] = raw_web_app.site_config
        web_app_dict['traffic_manager_host_names'] = raw_web_app.traffic_manager_host_names
        web_app_dict['scm_site_also_stopped'] = raw_web_app.scm_site_also_stopped
        web_app_dict['target_swap_slot'] = raw_web_app.target_swap_slot
        web_app_dict['hosting_environment_profile'] = raw_web_app.hosting_environment_profile
        web_app_dict['client_affinity_enabled'] = raw_web_app.client_affinity_enabled
        web_app_dict['client_cert_enabled'] = raw_web_app.client_cert_enabled
        web_app_dict['client_cert_exclusion_paths'] = raw_web_app.client_cert_exclusion_paths
        web_app_dict['host_names_disabled'] = raw_web_app.host_names_disabled
        web_app_dict['outbound_ip_addresses'] = raw_web_app.outbound_ip_addresses
        web_app_dict['possible_outbound_ip_addresses'] = raw_web_app.possible_outbound_ip_addresses
        web_app_dict['container_size'] = raw_web_app.container_size
        web_app_dict['daily_memory_time_quota'] = raw_web_app.daily_memory_time_quota
        web_app_dict['suspended_till'] = raw_web_app.suspended_till
        web_app_dict['max_number_of_workers'] = raw_web_app.max_number_of_workers
        web_app_dict['cloning_info'] = raw_web_app.cloning_info
        web_app_dict['resource_group'] = raw_web_app.resource_group
        web_app_dict['is_default_container'] = raw_web_app.is_default_container
        web_app_dict['default_host_name'] = raw_web_app.default_host_name
        web_app_dict['slot_swap_status'] = raw_web_app.slot_swap_status
        web_app_dict['redundancy_mode'] = raw_web_app.redundancy_mode
        web_app_dict['in_progress_operation_id'] = raw_web_app.in_progress_operation_id
        web_app_dict['identity'] = raw_web_app.identity
        web_app_dict['additional_properties'] = raw_web_app.additional_properties

        if raw_web_app.config is not None:
            web_app_dict['minimum_tls_version_supported'] = raw_web_app.config.min_tls_version
            web_app_dict['http_2_enabled'] = raw_web_app.config.http20_enabled

            # TODO handle this
            if raw_web_app.config.net_framework_version:
                web_app_dict['programming_language'] = 'dotnet'
                web_app_dict['programming_language_version'] = raw_web_app.config.net_framework_version
            elif raw_web_app.config.php_version:
                web_app_dict['programming_language'] = 'php'
                web_app_dict['programming_language_version'] = raw_web_app.config.php_version
            elif raw_web_app.config.python_version:
                web_app_dict['programming_language'] = 'python'
                web_app_dict['programming_language_version'] = raw_web_app.config.python_version
            elif raw_web_app.config.node_version:
                web_app_dict['programming_language'] = 'node'
                web_app_dict['programming_language_version'] = raw_web_app.config.node_version
            elif raw_web_app.config.java_version:
                web_app_dict['programming_language'] = 'java'
                web_app_dict['programming_language_version'] = raw_web_app.config.java_version
            else:
                web_app_dict['programming_language'] = None
                web_app_dict['programming_language_version'] = None

            # TODO - write rules for this values
            # web_app_dict[''] = raw_web_app.config.cors
            # web_app_dict[''] = raw_web_app.config.http_logging_enabled
            # web_app_dict[''] = raw_web_app.config.ftps_state
            # also look at network configuration / IP security restrictions
        else:
            web_app_dict['minimum_tls_version_supported'] = None
            web_app_dict['http_2_enabled'] = None
            web_app_dict['programming_language'] = None
            web_app_dict['programming_language_version'] = None

        if raw_web_app.auth_settings is not None:
            web_app_dict['authentication_enabled'] = raw_web_app.auth_settings.enabled
        else:
            web_app_dict['authentication_enabled'] = None

        return web_app_dict['id'], web_app_dict



from ScoutSuite.providers.azure.facade.base import AzureFacade
from ScoutSuite.providers.azure.resources.base import AzureResources
from ScoutSuite.providers.utils import get_non_provider_id


class Images(AzureResources):

    def __init__(self, facade: AzureFacade, subscription_id: str):
        super(Images, self).__init__(facade)
        self.subscription_id = subscription_id

    async def fetch_all(self):
        for raw_image in await self.facade.virtualmachines.get_images(self.subscription_id):
            id, image = self._parse_image(raw_image)
            self[id] = image

    def _parse_image(self, raw_image):
        # TODO
        pass



from ScoutSuite.providers.aws.facade.base import AWSFacade
from ScoutSuite.providers.aws.resources.base import AWSCompositeResources
from ScoutSuite.providers.utils import get_non_provider_id
from .grants import Grants


class Keys(AWSCompositeResources):
    _children = [
        (Grants, 'grants'),
    ]

    def __init__(self, facade: AWSFacade, region: str):
        super(Keys, self).__init__(facade)
        self.region = region

    async def fetch_all(self):
        raw_keys = await self.facade.kms.get_keys(self.region)
        for raw_key in raw_keys:
            key_id, key = self._parse_key(raw_key)
            self[key_id] = key

        await self._fetch_children_of_all_resources(
            resources=self,
            scopes={key_id: {'region': self.region, 'key_id': key['id']}
                    for (key_id, key) in self.items()}
        )

    def _parse_key(self, raw_key):
        key_dict = {}
        key_dict['id'] = key_dict['name'] = raw_key.get('KeyId')
        key_dict['arn'] = raw_key.get('KeyArn')
        key_dict['rotation_enabled'] = raw_key['rotation_status']['KeyRotationEnabled'] \
            if 'rotation_status' in raw_key else None

        key_dict['policy'] = raw_key.get('policy')

        if 'metadata' in raw_key:
            key_dict['creation_date'] = raw_key['metadata']['KeyMetadata']['CreationDate'] if \
                raw_key['metadata']['KeyMetadata']['CreationDate'] else None
            key_dict['key_enabled'] = False if raw_key['metadata']['KeyMetadata']['KeyState'] == 'Disabled' else True
            key_dict['description'] = raw_key['metadata']['KeyMetadata']['Description'] if len(
                raw_key['metadata']['KeyMetadata']['Description'].strip()) > 0 else None
            key_dict['origin'] = raw_key['metadata']['KeyMetadata']['Origin'] if len(
                raw_key['metadata']['KeyMetadata']['Origin'].strip()) > 0 else None
            key_dict['key_manager'] = raw_key['metadata']['KeyMetadata']['KeyManager'] if len(
                raw_key['metadata']['KeyMetadata']['KeyManager'].strip()) > 0 else None

        key_dict['aliases'] = {}
        for raw_alias in raw_key.get('aliases', []):
            alias_id, alias = self._parse_alias(raw_alias)
            key_dict['aliases'][alias_id] = alias

        return key_dict['id'], key_dict

    def _parse_alias(self, raw_alias):
        alias_dict = {
            # all KMS Aliases are prefixed with alias/, so we'll strip that off
            'id': get_non_provider_id(raw_alias.get('AliasArn')),
            'name': raw_alias.get('AliasName').split('alias/', 1)[-1],
            'arn': raw_alias.get('AliasArn'),
            'key_id': raw_alias.get('TargetKeyId')}
        return alias_dict['id'], alias_dict



import os

from ScoutSuite.core.console import print_exception

from ScoutSuite.providers.base.provider import BaseProvider
from ScoutSuite.providers.azure.services import AzureServicesConfig


class AzureProvider(BaseProvider):
    """
    Implements provider for Azure
    """

    def __init__(self,
                 subscription_ids=[], all_subscriptions=None,
                 report_dir=None, timestamp=None, services=None, skipped_services=None,
                 result_format='json',
                 **kwargs):
        services = [] if services is None else services
        skipped_services = [] if skipped_services is None else skipped_services

        self.metadata_path = '%s/metadata.json' % os.path.split(os.path.abspath(__file__))[0]

        self.provider_code = 'azure'
        self.provider_name = 'Microsoft Azure'
        self.environment = 'default'

        self.programmatic_execution = kwargs['programmatic_execution']
        self.credentials = kwargs['credentials']

        if subscription_ids:
            self.subscription_ids = subscription_ids
        elif self.credentials.default_subscription_id:
            self.subscription_ids = [self.credentials.default_subscription_id]
        else:
            self.subscription_ids = []
        self.all_subscriptions = all_subscriptions

        try:
            self.account_id = self.credentials.get_tenant_id()
        except Exception as e:
            self.account_id = 'undefined'

        self.services = AzureServicesConfig(self.credentials,
                                            programmatic_execution=self.programmatic_execution,
                                            subscription_ids=self.subscription_ids,
                                            all_subscriptions=self.all_subscriptions)

        self.result_format = result_format

        super(AzureProvider, self).__init__(report_dir, timestamp,
                                            services, skipped_services, result_format)

    def get_report_name(self):
        """
        Returns the name of the report using the provider's configuration
        """
        try:
            return 'azure-tenant-{}'.format(self.credentials.get_tenant_id())
        except Exception as e:
            print_exception('Unable to define report name: {}'.format(e))
            return 'azure'

    def preprocessing(self, ip_ranges=None, ip_ranges_name_key=None):
        """
        Tweak the Azure config to match cross-service resources and clean any fetching artifacts

        :param ip_ranges:
        :param ip_ranges_name_key:
        :return: None
        """
        ip_ranges = [] if ip_ranges is None else ip_ranges

        self._match_arm_roles_and_principals()

        super(AzureProvider, self).preprocessing()

    def _match_arm_roles_and_principals(self):
        """
        Matches ARM roles to AAD service principals

        :return:
        """

        # Add role assignments
        if 'arm' in self.service_list and 'aad' in self.service_list:
            for subscription in self.services['arm']['subscriptions']:
                for assignment in self.services['arm']['subscriptions'][subscription]['role_assignments'].values():
                    role_id = assignment['role_definition_id'].split('/')[-1]
                    for group in self.services['aad']['groups']:
                        if group == assignment['principal_id']:
                            self.services['aad']['groups'][group]['roles'].append({'subscription_id': subscription,
                                                                                 'role_id': role_id})
                            self.services['arm']['subscriptions'][subscription]['roles'][role_id]['assignments']['groups'].append(group)
                            self.services['arm']['subscriptions'][subscription]['roles'][role_id]['assignments_count'] += 1
                    for user in self.services['aad']['users']:
                        if user == assignment['principal_id']:
                            self.services['aad']['users'][user]['roles'].append({'subscription_id': subscription,
                                                                                 'role_id': role_id})
                            self.services['arm']['subscriptions'][subscription]['roles'][role_id]['assignments']['users'].append(user)
                            self.services['arm']['subscriptions'][subscription]['roles'][role_id]['assignments_count'] += 1
                    for service_principal in self.services['aad']['service_principals']:
                        if service_principal == assignment['principal_id']:
                            self.services['aad']['service_principals'][service_principal]['roles'].append({'subscription_id': subscription,
                                                                                                           'role_id': role_id})
                            self.services['arm']['subscriptions'][subscription]['roles'][role_id]['assignments']['service_principals'].append(service_principal)
                            self.services['arm']['subscriptions'][subscription]['roles'][role_id]['assignments_count'] += 1



from azure.mgmt.web import WebSiteManagementClient

from ScoutSuite.core.console import print_exception
from ScoutSuite.providers.azure.utils import get_resource_group_name
from ScoutSuite.providers.utils import run_concurrently, get_and_set_concurrently


class AppServiceFacade:

    def __init__(self, credentials):
        self.credentials = credentials

    def get_client(self, subscription_id: str):
        return WebSiteManagementClient(self.credentials.arm_credentials, subscription_id=subscription_id)

    async def get_web_apps(self, subscription_id: str):
        try:
            client = self.get_client(subscription_id)
            web_apps = await run_concurrently(
                lambda: list(client.web_apps.list())
            )
        except Exception as e:
            print_exception('Failed to retrieve web apps: {}'.format(e))
            return []
        else:
            await get_and_set_concurrently([self._get_and_set_web_app_configuration], web_apps, api_client=client)
            await get_and_set_concurrently([self._get_and_set_web_app_auth_settings], web_apps, api_client=client)
            return web_apps

    async def _get_and_set_web_app_configuration(self, web_app, api_client):
        resource_group_name = get_resource_group_name(web_app.id)
        try:
            web_app_config = await run_concurrently(
                lambda: api_client.web_apps.get_configuration(resource_group_name, web_app.name)
            )
        except Exception as e:
            print_exception('Failed to retrieve web app configuration: {}'.format(e))
            setattr(web_app, 'config', None)
        else:
            setattr(web_app, 'config', web_app_config)

    async def _get_and_set_web_app_auth_settings(self, web_app, api_client):
        resource_group_name = get_resource_group_name(web_app.id)
        try:
            web_app_auth_settings = await run_concurrently(
                lambda: api_client.web_apps.get_auth_settings(resource_group_name=resource_group_name,
                                                              name=web_app.name)
            )
        except Exception as e:
            print_exception('Failed to retrieve web app auth settings: {}'.format(e))
            setattr(web_app, 'auth_settings', None)
        else:
            setattr(web_app, 'auth_settings', web_app_auth_settings)



from ScoutSuite.providers.azure.resources.subscriptions import Subscriptions

from .role_assignments import RoleAssignments
from .roles import Roles


class ARM(Subscriptions):
    _children = [
        (Roles, 'roles'),
        (RoleAssignments, 'role_assignments')
    ]



from ScoutSuite.providers.azure.resources.base import AzureCompositeResources

from .users import Users
from .groups import Groups
from .serviceprincipals import ServicePrincipals
from .applications import Applications


class AAD(AzureCompositeResources):
    _children = [
        (Users, 'users'),
        (Groups, 'groups'),
        (ServicePrincipals, 'service_principals'),
        (Applications, 'applications'),
    ]

    async def fetch_all(self):
        await self._fetch_children(resource_parent=self)

    async def finalize(self):

        # Add group members
        for group in self['groups']:
            for user in self['users']:
                if group in self['users'][user]['groups']:
                    self['groups'][group]['users'].append(user)



from ScoutSuite.providers.aws.facade.base import AWSFacade
from ScoutSuite.providers.aws.resources.base import AWSResources


class Grants(AWSResources):
    def __init__(self, facade: AWSFacade, region: str, key_id: str):
        super(Grants, self).__init__(facade)
        self.region = region
        self.key_id = key_id

    async def fetch_all(self):
        raw_grants = await self.facade.kms.get_grants(self.region, self.key_id)
        for raw_grant in raw_grants:
            id, grant = self._parse_grant(raw_grant)
            self[id] = grant

    def _parse_grant(self, raw_grant):
        grant_dict = {
            'key_id': raw_grant.get('KeyId'),
            'grant_id': raw_grant.get('GrantId'),
            'name': raw_grant.get('Name'),
            'create_date': raw_grant.get('CreationDate'),
            'grantee_principal': raw_grant.get('GranteePrincipal'),
            'retiring_principal': raw_grant.get('ReitirngPrincipal'),
            'issuing_account': raw_grant.get('IssuingAccount'),
            'operations': raw_grant.get('Operations'),
            'constraints': raw_grant.get('Constraints')
        }
        return grant_dict['grant_id'], grant_dict



from ScoutSuite.providers.azure.facade.base import AzureFacade
from ScoutSuite.providers.azure.resources.base import AzureResources


class Alerts(AzureResources):

    def __init__(self, facade: AzureFacade, subscription_id: str):
        super(Alerts, self).__init__(facade)
        self.subscription_id = subscription_id

    async def fetch_all(self):
        a = await self.facade.securitycenter.get_alerts(self.subscription_id)
        for raw_alert in await self.facade.securitycenter.get_alerts(self.subscription_id):
            id, alert = self._parse_alert(raw_alert)
            self[id] = alert

    def _parse_alert(self, alert):
        alert_dict = {}
        alert_dict['id'] = alert.id
        alert_dict['name'] = alert.name
        return alert_dict['id'], alert_dict



from azure.mgmt.keyvault import KeyVaultManagementClient

from ScoutSuite.core.console import print_exception
from ScoutSuite.providers.utils import run_concurrently


class KeyVaultFacade:
    def __init__(self, credentials):
        self.credentials = credentials

    def get_client(self, subscription_id: str):
        return KeyVaultManagementClient(self.credentials.arm_credentials, subscription_id=subscription_id)

    async def get_key_vaults(self, subscription_id: str):
        try:
            client = self.get_client(subscription_id)
            return await run_concurrently(
                lambda: list(client.vaults.list_by_subscription()))
        except Exception as e:
            print_exception('Failed to retrieve key vaults: {}'.format(e))
            return []
